EXPORTS
	encode_HDB_keyset
	_hdb_fetch_kvno
	_hdb_remove
	_hdb_store
	hdb_add_current_keys_to_history
	hdb_add_history_key
	hdb_add_history_keyset
	hdb_add_master_key
        hdb_change_kvno
	hdb_check_db_format
	hdb_clear_extension
	hdb_clear_master_key
	hdb_create
	hdb_db_dir
	hdb_dbinfo_get_acl_file
	hdb_dbinfo_get_binding
	hdb_dbinfo_get_dbname
	hdb_dbinfo_get_label
	hdb_dbinfo_get_log_file
	hdb_dbinfo_get_mkey_file
	hdb_dbinfo_get_next
	hdb_dbinfo_get_realm
	hdb_derive_etypes
	hdb_default_db
	hdb_enctype2key
	hdb_entry2string
	hdb_entry2value
	hdb_entry_add_key_rotation
	hdb_entry_alias2value
	hdb_entry_check_mandatory
	hdb_entry_clear_password
	hdb_entry_get_ConstrainedDelegACL
	hdb_entry_get_aliases
	hdb_entry_get_key_rotation
	hdb_entry_get_krb5_config
	hdb_entry_get_password
	hdb_entry_get_pkinit_acl
	hdb_entry_get_pkinit_cert
	hdb_entry_get_pkinit_hash
	hdb_entry_get_pw_change_time
	hdb_entry_set_krb5_config
	hdb_entry_set_password
	hdb_entry_set_pw_change_time
	hdb_fetch_kvno
	hdb_find_extension
	hdb_foreach
	hdb_free_dbinfo
	hdb_free_entry
	hdb_free_key
	hdb_free_keys
	hdb_free_master_key
	hdb_generate_key_set
	hdb_generate_key_set_password
	hdb_generate_key_set_password_with_ks_tuple
	hdb_get_dbinfo
	hdb_get_instance
	hdb_init_db
	hdb_install_keyset
        hdb_interface_version   DATA
	hdb_key2principal
	hdb_kvno2keys
	hdb_list_builtin
	hdb_lock
	hdb_next_enctype2key
	hdb_principal2key
	hdb_print_entry
	hdb_process_master_key
	hdb_prune_keys
	hdb_prune_keys_kvno
	hdb_read_master_key
	hdb_remove_keys
	hdb_replace_extension
	hdb_seal_key
	hdb_seal_key_mkey
	hdb_seal_keys
	hdb_seal_keys_mkey
        hdb_set_last_modified_by
	hdb_set_master_key
	hdb_set_master_keyfile
	hdb_unlock
	hdb_unseal_key
	hdb_unseal_key_mkey
	hdb_unseal_keys
	hdb_unseal_keys_kvno
	hdb_unseal_keys_mkey
	hdb_validate_key_rotation
	hdb_validate_key_rotations
	hdb_value2entry
	hdb_value2entry_alias
	hdb_write_master_key
	length_HDB_keyset
        initialize_hdb_error_table_r

	hdb_kt_ops
	hdb_get_kt_ops

; MIT KDB related entries
	_hdb_mdb_value2entry
	_hdb_mit_dump2mitdb_entry

; some random bits needed for libkadm
	add_HDB_Ext_KeyRotation
	add_HDB_Ext_KeySet
	add_Keys
	asn1_HDBFlags_units
	copy_Event
	copy_HDB_EncTypeList
	copy_hdb_entry
	copy_HDB_entry
	copy_hdb_entry_alias
        copy_HDB_entry_alias
	copy_HDB_EntryOrAlias
	copy_HDB_extensions
	copy_HDB_Ext_KeyRotation
	copy_Key
	copy_Keys
	copy_Salt
	decode_HDB_EncTypeList
	decode_hdb_entry
	decode_HDB_entry
	decode_hdb_entry_alias
        decode_HDB_entry_alias
	decode_HDB_EntryOrAlias
	decode_HDB_Ext_Aliases
	decode_HDB_extension
	decode_HDB_Ext_KeyRotation
	decode_HDB_Ext_PKINIT_acl
	decode_Key
	decode_Keys
	encode_HDB_EncTypeList
	encode_hdb_entry
	encode_HDB_entry
	encode_hdb_entry_alias
	encode_HDB_entry_alias
	encode_HDB_EntryOrAlias
	encode_HDB_Ext_Aliases
	encode_HDB_extension
	encode_HDB_Ext_KeyRotation
	encode_HDB_Ext_PKINIT_acl
	encode_hdb_keyset
	encode_Key
	encode_Keys
	free_Event
	free_HDB_EncTypeList
	free_hdb_entry
	free_HDB_entry
	free_hdb_entry_alias
	free_HDB_entry_alias
	free_HDB_EntryOrAlias
	free_HDB_Ext_Aliases
	free_HDB_extension
	free_HDB_extensions
	free_HDB_Ext_KeyRotation
	free_HDB_Ext_KeySet
	free_HDB_Ext_PKINIT_acl
	free_hdb_keyset
	free_HDB_keyset
	free_Key
	free_Keys
	free_Salt
	HDBFlags2int
	int2HDBFlags
	int2KeyRotationFlags
	KeyRotationFlags2int
	length_HDB_EncTypeList
	length_hdb_entry
	length_HDB_entry
	length_hdb_entry_alias
	length_HDB_entry_alias
	length_HDB_EntryOrAlias
	length_HDB_Ext_Aliases
	length_HDB_extension
	length_HDB_Ext_KeyRotation
	length_HDB_Ext_PKINIT_acl
	length_hdb_keyset
	length_Key
	length_Keys
	remove_HDB_Ext_KeyRotation
	remove_Keys
